from dataclasses import dataclass, field


@dataclass
class BaseStructure:
    @dataclass
    class ProjectStructure:
        name: str
        version: str
        description: str
        readme: str

        @dataclass
        class AuthorStructure:
            name: str
            email: str

        authors: list[AuthorStructure]
        requires_python: str = field(metadata={"name": "requires-python"})
        dependencies: list

    project: ProjectStructure

    @dataclass
    class BuildSystemStructure:
        requires: list[str]
        build_backend: str = field(metadata={"name": "build-backend"})

    build_system: BuildSystemStructure = field(metadata={"name": "build-system"})


EXAMPLE_TOML = """
[project]
name = "tattl"
version = "dev"
description = "Totally Awesome Type-aware TOML Loader"
readme = "README.md"
authors = [
    { name = "thcrt", email = "110127860+thcrt@users.noreply.github.com" }
]
requires-python = ">=3.12"
dependencies = []

[build-system]
requires = ["hatchling"]
build-backend = "hatchling.build"

[dependency-groups]
dev = [
    "mypy>=1.13.0",
    "pytest>=8.3.4",
    "ruff>=0.8.2",
]
"""


def test_example():
    import tattl
    import tomllib

    data = tattl.unpack_dict(tomllib.loads(EXAMPLE_TOML), BaseStructure)
    print(data)
    assert data == BaseStructure(
        project=BaseStructure.ProjectStructure(
            name="tattl",
            version="dev",
            description="Totally Awesome Type-aware TOML Loader",
            readme="README.md",
            authors=[
                BaseStructure.ProjectStructure.AuthorStructure(
                    name="thcrt", email="110127860+thcrt@users.noreply.github.com"
                )
            ],
            requires_python=">=3.12",
            dependencies=[],
        ),
        build_system=BaseStructure.BuildSystemStructure(
            requires=["hatchling"], build_backend="hatchling.build"
        ),
    )



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import tattl
import datetime

project = "tattl"
author = "thcrt"
copyright = f"{datetime.date.today().year}, thcrt"
version = tattl.__version__
release = tattl.__version__

needs_sphinx = "8.1"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx_paramlinks",
    "sphinx.ext.autodoc.typehints",
    "sphinx_multiversion",
]

highlight_language = "python3"

nitpicky = True

maximum_signature_line_length = 80

templates_path = ["_templates"]
exclude_patterns = []

html_theme = "furo"
html_theme_options = {
    "source_repository": "https://github.com/thcrt/tattl/",
    "source_branch": "main",
    "source_directory": "docs/",
    "footer_icons": [
        {
            "name": "GitHub",
            "url": "https://github.com/thcrt/tattl",
            "html": """
                <svg stroke="currentColor" fill="currentColor" stroke-width="0" viewBox="0 0 16 16">
                    <path fill-rule="evenodd" d="M8 0C3.58 0 0 3.58 0 8c0 3.54 2.29 6.53 5.47 7.59.4.07.55-.17.55-.38 0-.19-.01-.82-.01-1.49-2.01.37-2.53-.49-2.69-.94-.09-.23-.48-.94-.82-1.13-.28-.15-.68-.52-.01-.53.63-.01 1.08.58 1.23.82.72 1.21 1.87.87 2.33.66.07-.52.28-.87.51-1.07-1.78-.2-3.64-.89-3.64-3.95 0-.87.31-1.59.82-2.15-.08-.2-.36-1.02.08-2.12 0 0 .67-.21 2.2.82.64-.18 1.32-.27 2-.27.68 0 1.36.09 2 .27 1.53-1.04 2.2-.82 2.2-.82.44 1.1.16 1.92.08 2.12.51.56.82 1.27.82 2.15 0 3.07-1.87 3.75-3.65 3.95.29.25.54.73.54 1.48 0 1.07-.01 1.93-.01 2.2 0 .21.15.46.55.38A8.013 8.013 0 0 0 16 8c0-4.42-3.58-8-8-8z"></path>
                </svg>
            """,
            "class": "",
        },
    ],
}

html_title = f"{project} docs"
html_last_updated_fmt = "%b %d, %Y"

html_static_path = ["_static"]
html_css_files = ["css/versions.css"]
html_js_files = ["js/versions.js"]

smv_released_pattern = r"^refs/tags/.*$"

add_function_parentheses = False

intersphinx_mapping = {"python": ("https://docs.python.org/3", None)}

autodoc_class_signature = "separated"
paramlinks_hyperlink_param = "name"


def remove_module_docstring(app, what, name, obj, options, lines):
    if what == "module" and name == "tattl":
        del lines[:]


def setup(app):
    app.connect("autodoc-process-docstring", remove_module_docstring)



"""Totally Awesome Type-Aware TOML Loader

TATTL is a library to deserialize data, particularly application configuration, while validating
types. It builds an instance of a given ``dataclass`` and checks the value of each field against its
type annotations. It supports nested ``dataclass`` structures and modern type definitions.

This is the base module for ``tattl``, and the package's functionality is available under its
various submodules.

Licensed under the BSD 3-Clause License. Copyright Theo Court and contributors.

https://github.com/thcrt/tattl


"""

from .unpack import unpack_dict

__all__ = ["unpack_dict", "__version__"]


# Define the variable '__version__':
try:
    # If setuptools_scm is installed (e.g. in a development environment with
    # an editable install), then use it to determine the version dynamically.
    from setuptools_scm import get_version as _get_version  # type: ignore

    # This will fail with LookupError if the package is not installed in
    # editable mode or if Git is not installed.
    __version__ = _get_version(root="..", relative_to=__file__)
except (ImportError, LookupError):
    # As a fallback, use the version that is hard-coded in the file.
    try:
        from ._version import __version__  # noqa: F401
    except ModuleNotFoundError:
        # The user is probably trying to run this without having installed
        # the package, so complain.
        raise RuntimeError(
            "tattl is not correctly installed. Install it with pip or uv."
        )



from dataclasses import dataclass, field


def my_dict_factory():
    return {"spam": 1.23, "ham": 4.56, "eggs": 7.89}


def my_list_factory():
    return [1, 1, 2, 3, 5, 8]


@dataclass
class Structure:
    foo: str = "AAA"
    bar: int = 123
    baz: bool = field(default=True)
    qux: list[str] = field(default_factory=list)
    cor: dict[str, float] = field(default_factory=my_dict_factory)
    gra: list[int] = field(default_factory=my_list_factory)


EXAMPLE_TOML = """
# There's nothing here! Hopefully we see our defaults in use.
"""


def test_example():
    import tattl
    import tomllib

    data = tattl.unpack_dict(tomllib.loads(EXAMPLE_TOML), Structure)
    print(data)
    assert data == Structure(
        foo="AAA",
        bar=123,
        baz=True,
        qux=[],
        cor={"spam": 1.23, "ham": 4.56, "eggs": 7.89},
        gra=[1, 1, 2, 3, 5, 8],
    )



from dataclasses import dataclass
import datetime


@dataclass
class BaseStructure:
    title: str

    @dataclass
    class OwnerStructure:
        name: str
        dob: datetime.datetime

    owner: OwnerStructure

    @dataclass
    class DatabaseStructure:
        enabled: bool
        ports: list[int]
        data: list[list[str | float]]
        temp_targets: dict[str, float]

    database: DatabaseStructure

    @dataclass
    class ServerStructure:
        ip: str
        role: str

    servers: dict[str, ServerStructure]


EXAMPLE_TOML = """
# This is a TOML document

title = "TOML Example"

[owner]
name = "Tom Preston-Werner"
dob = 1979-05-27T07:32:00-08:00

[database]
enabled = true
ports = [ 8000, 8001, 8002 ]
data = [ ["delta", "phi"], [3.14] ]
temp_targets = { cpu = 79.5, case = 72.0 }

[servers]

[servers.alpha]
ip = "10.0.0.1"
role = "frontend"

[servers.beta]
ip = "10.0.0.2"
role = "backend"
"""


def test_example():
    import tattl
    import tomllib

    data = tattl.unpack_dict(tomllib.loads(EXAMPLE_TOML), BaseStructure)
    print(data)
    assert data == BaseStructure(
        title="TOML Example",
        owner=BaseStructure.OwnerStructure(
            name="Tom Preston-Werner",
            dob=datetime.datetime(
                1979,
                5,
                27,
                7,
                32,
                tzinfo=datetime.timezone(datetime.timedelta(days=-1, seconds=57600)),
            ),
        ),
        database=BaseStructure.DatabaseStructure(
            enabled=True,
            ports=[8000, 8001, 8002],
            data=[["delta", "phi"], [3.14]],
            temp_targets={"cpu": 79.5, "case": 72.0},
        ),
        servers={
            "alpha": BaseStructure.ServerStructure(ip="10.0.0.1", role="frontend"),
            "beta": BaseStructure.ServerStructure(ip="10.0.0.2", role="backend"),
        },
    )



from dataclasses import dataclass, field


def my_dict_factory():
    return {"spam": 1.23, "ham": 4.56, "eggs": 7.89}


def my_list_factory():
    return [1, 1, 2, 3, 5, 8]


@dataclass
class Structure:
    foo: str = "AAA"
    bar: int = 123
    baz: bool = field(default=True)
    qux: list[str] = field(default_factory=list)
    cor: dict[str, float] = field(default_factory=my_dict_factory)
    gra: list[int] = field(default_factory=my_list_factory)


EXAMPLE_TOML = """
foo = "BBB"
bar = 234
baz = false
qux = [ "Hello", "World" ]
cor = { snork = 2.34, blarg = 5.67, wibble = 8.90 }
gra = [ 1, 121, 12321, 1234321 ]
"""


def test_example():
    import tattl
    import tomllib

    data = tattl.unpack_dict(tomllib.loads(EXAMPLE_TOML), Structure)
    print(data)
    assert data == Structure(
        foo="BBB",
        bar=234,
        baz=False,
        qux=["Hello", "World"],
        cor={"snork": 2.34, "blarg": 5.67, "wibble": 8.90},
        gra=[1, 121, 12321, 1234321],
    )



class TATTLException(Exception):
    """Raised when the data given does not conform to the ``dataclass``.

    This is the base exception for all TATTL exceptions."""


class ValidationException(TATTLException):
    """Indicates that the type of a field in the data differs from the type of the ``dataclass``
    field with the same name.
    """

    pass


class MissingFieldException(TATTLException):
    """Indicates that the data is missing a required field in the ``dataclass``."""
